# Recursive formulation
def C(n,k):
    """
    Counts how many different subsets of size 
    k there are in a set with n elements
    """
    if (n==k or k==0):
        return 1 
    else:
        return C(n-1,k-1)+C(n-1,k)
    
# Dynamic programming formulation
import numpy as np 
def C_dp(n,k):
    """
    Counts how many different subsets of size 
    k there are in a set with n elements
    """
    
    # Use 64bit integer precision for the array entries
    C = np.zeros((k+1,n+1), dtype=np.int64) 

    # Initialize base cases along in first row
    for j in range(n+1):
        C[0,j]=1

    # Initialize base cases in the diagonal
    for j in range(min(n+1,k+1)):
        C[j,j]=1

    # Compute the solution to the rest of cases 
    # in the appropriate order to satisfy the dependencies
    # among subproblems
    for i in range(1,k+1):
        for j in range(i+1,n+1):
            C[i,j]=C[i-1,j-1]+C[i,j-1]

    return C[k,n] 

%timeit C(100,5)
# WARNING: wery time-consuming, might need to interrupt the kernel
%timeit C(100,6)

%timeit C_dp(100,5)
%timeit C_dp(100,6)


# Recursive formulation
def edit_distance(s,t,w_sub=1,w_gap=1):
    """
    Returns the edit distance between 2 sequences
    s and t with gap cost w gap and substitution
    cost w sub
    """

    if (len(s)==0):
        return len(t)*w_gap
    elif (len(t)==0):
        return len(s)*w_gap
    
    # Possibility 1 
    if s[-1]==t[-1]:
        d1=edit_distance(s[:-1],t[:-1],w_sub,w_gap)
    else:
        d1=edit_distance(s[:-1],t[:-1],w_sub,w_gap)+w_sub
    
    # Possibility 2 
    d2=edit_distance(s[:-1],t,w_sub,w_gap)+w_gap

    # Possibility 3
    d3=edit_distance(s,t[:-1],w_sub,w_gap)+w_gap

    return min(d1,d2,d3)

# Dynamic programming formulation
import numpy as np 
def edit_distance_dp(s,t,w_sub=1,w_gap=1):
    """
    Returns the edit distance between 2 sequences
    s and t with gap cost w gap and substitution
    cost w sub
    """

    # Use 64bit integer precision for the array entries
    dist = np.zeros((len(s)+1,len(t)+1), dtype=np.int64)

    # Initialize base cases in first column
    for i in range(dist.shape[0]):
        dist[i,0]=i

    # Initialize base cases in second column
    for j in range(dist.shape[1]):
        dist[0,j]=j

    # Compute the rest of cases 
    for i in range(1,dist.shape[0]):
        for j in range(1,dist.shape[1]):
            d1=dist[i-1,j-1]+int(s[i-1]!=t[j-1])*w_sub
            d2=dist[i,j-1]+w_gap
            d3=dist[i-1,j]+w_gap
            dist[i,j]=min(d1,d2,d3)

    return dist[-1,-1]

%timeit edit_distance("TGCTA","GCATA")
# WARNING: wery time-consuming, might need to interrupt the kernel
%timeit edit_distance("GAATTCAAAATTT","GGATCGAGGGAAA") 

%timeit edit_distance_dp("TGCTA","GCATA")
%timeit edit_distance_dp("GAATTCAAAATTT","GGATCGAGGGAAA")



